// Copyright 2000 Ogasawara Hiroyuki
// 2000/10/09

#include	<linux/config.h>
#include	<linux/mm.h>
#include	<linux/malloc.h>
#include	<linux/vmalloc.h>
#include	<linux/init.h>
#include	<asm/uaccess.h>

#include	<devmodel.h>

extern unsigned short	*GetVramTop(void);
extern unsigned char	EnableAccessVRAM(void);
extern void		DisableAccessVRAM( unsigned char );

//#include	"../../_debugcode.h"


typedef struct {
	unsigned short	Width;
	unsigned short	Height;
	unsigned short	WLine;
	unsigned short	BitCount;
	ssize_t		(*exec)( const char *ptr, size_t count, loff_t* );
} _WriteExec;

static _WriteExec	*CurSelect= NULL;

static void
_CopyToMono( unsigned char *str, const unsigned short *ptr, int len )
{
	unsigned short	*sp= (unsigned short*)
				(((unsigned int)str) & 0xfffffffe);
	for( len>>= 2 ; len-- ;){
		unsigned int	c, r, g, b, rr;

		c= *ptr++;
		r= (c >> (6+5-3  +3  )) & 0x1f;
		g= (c >> (  5-2-1+3  )) & 0x3f;
		b= (c                 ) & 0x1f;
		rr= (((g*3+r*3+b)*100)<<( 2-1)) & 0xf000;

		c= *ptr++;
		r= (c >> (6+5-3  +3  )) & 0x1f;
		g= (c >> (  5-2-1+3  )) & 0x3f;
		b= (c                 ) & 0x1f;
		rr|= (((g*3+r*3+b)*100)>>( 2+1)) & 0x0f00;

		c= *ptr++;
		r= (c >> (6+5-3  +3  )) & 0x1f;
		g= (c >> (  5-2-1+3  )) & 0x3f;
		b= (c                 ) & 0x1f;
		rr|= (((g*3+r*3+b)*100)>>( 6+1)) & 0x00f0;

		c= *ptr++;
		r= (c >> (6+5-3  +3  )) & 0x1f;
		g= (c >> (  5-2-1+3  )) & 0x3f;
		b= (c                 ) & 0x1f;
		rr|= (((g*3+r*3+b)*100)>>(10+1)) & 0x000f;

		*sp++= ~rr;
	}
}

static ssize_t
WriteQVGAM( const char *ptr, size_t count, loff_t *ppos )
{
	unsigned long	pos= *ppos>>1;
	unsigned char	acc= EnableAccessVRAM();
	unsigned char	*top= (unsigned char*)GetVramTop();
	unsigned short	LineBuf[ 512 ];

	int		x, y, line;
	int		err= count;
	unsigned char	*st;

	*ppos+= count;

	x= pos % CurSelect->Width;
	y= pos / CurSelect->Width;
	line= (CurSelect->Width-x) * sizeof(short);
	if( line > count )
		line= count;
	st= top+ ((y*CurSelect->WLine + x)>>1);

	for(; count && y < CurSelect->Height ;){
		copy_from_user( LineBuf, ptr, line );
		_CopyToMono( st, LineBuf, line>>1 );
		ptr+= line;
		count-= line;
		line= count < CurSelect->Width*sizeof(short)
				? count : CurSelect->Width*sizeof(short);
		st= top + (++y)*(CurSelect->WLine>>1);
	}
	DisableAccessVRAM( acc );
	return	err;
}

static ssize_t
WriteVGA( const char *ptr, size_t count, loff_t *ppos )
{
	unsigned long	pos= *ppos>>1;
	unsigned char	acc= EnableAccessVRAM();
	unsigned short	*top= GetVramTop();

	int		x, y, line;
	int		err= count;
	unsigned short	*st;

	*ppos+= count;

	x= pos % CurSelect->Width;
	y= pos / CurSelect->Width;
	line= (CurSelect->Width-x) * sizeof(short);
	if( line > count )
		line= count;
	st= top+ y*CurSelect->WLine + x;

	for(; count && y < CurSelect->Height ;){
		copy_from_user( st, ptr, line );
		ptr+= line;
		count-= line;
		line= count < CurSelect->Width*sizeof(short)
				? count : CurSelect->Width*sizeof(short);
		st= top + (++y)*CurSelect->WLine;
	}
	DisableAccessVRAM( acc );
	return	err;
}

static ssize_t
WriteQVGA( const char *ptr, size_t count, loff_t *ppos )
{
	unsigned long	pos= *ppos;
	unsigned char	acc= EnableAccessVRAM();
	unsigned short	*top= GetVramTop();
	int		err= count;
	if( copy_from_user( ((char*)top) + pos, ptr, count ) ){
		err= -EFAULT;
	}
	*ppos+= count;
	DisableAccessVRAM( acc );
	return	err;
}

static _WriteExec	_WriteExecTable[]= {
{	320,	240,	320,	4,	WriteQVGAM,	},
{	320,	240,	320,	16,	WriteQVGA,	},
{	320,	240,	512,	4,	WriteQVGAM,	},
{	640,	480,	1024,	16,	WriteVGA,	},
};

static _WriteExec *
SelectType( void )
{
	unsigned char	sysModelName[22];
	unsigned long	SysVersion;
	GetSystemInfo( sysModelName, &SysVersion );

	switch( (SysVersion >>16) & 0xffff ){
#if 0
	case MODEL_NUMBER_MI10:	// 001	MI-10DC/MI-10
	case MODEL_NUMBER_BO185:// 00d	BO-185
	case MODEL_NUMBER_BIL10:// 00e	BI-L10
	case MODEL_NUMBER_MIP3: // 00f	MI-P3 (??)
#endif
	default:
		break;
	case 0x00c://MODEL_NUMBER_MIP10:// 00c	MI-P10 igeti
		return	&_WriteExecTable[2];
#if 0
	case 0x010://MODEL_NUMBER_MIJ1: // 010	MI-J1 igeti
	case MODEL_NUMBER_MT200:// 006	MT-200 ComminucationPal
	case 0x00a://MODEL_NUMBER_MT300:// 00a	MT-300 ComminucationPal
	case 0x009://MODEL_NUMBER_BO181:// 009	BO-181	BrowserBoard
	case MODEL_NUMBER_MI100:// 003	MI-110M/110/106
	case MODEL_NUMBER_MIP1:	// 007	MI-P1/P2 igeti
		return	&_WriteExecTable[0];
#endif
	case 0x00b://MODEL_NUMBER_MIC1:	// 00b	MI-C1
#if 0
	case MODEL_NUMBER_MI600:// 004	MI-610DC/610
	case MODEL_NUMBER_MI300:// 005	MI-310 ColorPocket
	case MODEL_NUMBER_MI50:	// 002	MI-506DC/506/504
#endif
		return	&_WriteExecTable[1];
	case MODEL_NUMBER_MIEX1:// 008	MI-EX1
		return	&_WriteExecTable[3];
	}
	return	NULL;
}

static ssize_t
zaugraph_common_write( struct file *file, const char *ptr,
						size_t count, loff_t *ppos )
{
	if( CurSelect ){
		return	(*CurSelect->exec)( ptr, count, ppos );
	}
	return	0;
}

static int
zaugraph_common_open( struct inode *inode, struct file *file )
{
	CurSelect= SelectType();
	return	0;
}



static ssize_t
zaugraph_raw_write( struct file *file, const char *ptr,
						size_t count, loff_t *ppos )
{
	return	WriteQVGA( ptr, count, ppos );
}


static loff_t
zaugraph_raw_lseek( struct file *file, loff_t offset, int dir )
{
	switch( dir ){
	case 0:
		file->f_pos= offset;
		return	file->f_pos;
	case 1:
		file->f_pos+= offset;
		return	file->f_pos;
	default:
		break;
	}
	return	-EINVAL;
}

static int
zaugraph_raw_ioctl( struct inode *inode, struct file *file,
				unsigned int cmd, unsigned long arg )
{
	switch( cmd ){
	case 0:
		if( CurSelect ){
			copy_to_user( (void*)arg, CurSelect,
						sizeof(_WriteExec) );
		}
		return	0;
	}
	return	-EINVAL;
}



static struct file_operations	zaugraph_common_fops= {
	zaugraph_raw_lseek,	// lseek
	NULL,//zaugraph_read,	// read
	zaugraph_common_write,	// write
	NULL,			// readdir
	NULL,			// poll
	zaugraph_raw_ioctl,	// ioctl
	NULL,			// mmap
	zaugraph_common_open,	// open
	NULL,			// flush
	NULL,			// release
	NULL			// fsync
};


static struct file_operations	zaugraph_raw_fops= {
	zaugraph_raw_lseek,	// lseek
	NULL,//zaugraph_read,	// read
	zaugraph_raw_write,	// write
	NULL,			// readdir
	NULL,			// poll
	zaugraph_raw_ioctl,	// ioctl
	NULL,			// mmap
	zaugraph_common_open,	// open
	NULL,			// flush
	NULL,			// release
	NULL			// fsync
};




static int
zaugraph_open( struct inode *inode, struct file *file )
{
	switch( MINOR( inode->i_rdev ) ){
	case 0:
		file->f_op= &zaugraph_common_fops;
		break;
	case 8:
		file->f_op= &zaugraph_raw_fops;
		break;
	case 1:
		//file->f_op= &zaugraph_commonvga_fops;
		//break;
	default:
		return	-ENXIO;
	}
	if( file->f_op && file->f_op->open ){
		return	(*file->f_op->open)( inode, file );
	}
	return	0;
}

static struct file_operations	zaugraph_fops= {
	NULL,		/* lseek */
	NULL,		/* read */
	NULL,		/* write */
	NULL,		/* readdir */
	NULL,		/* poll */
	NULL,		/* ioctl */
	NULL,		/* mmap */
	zaugraph_open,	/* open */
	NULL,		/* flush */
	NULL,		/* release */
	NULL		/* fsync */
};

int __init zaugraph_init( void )
{
	if( register_chrdev( 158, "zgraph", &zaugraph_fops ) ){
		printk("unable to get major 158 for zgraph devs\n" );
	}
	return	0;
}

