//	Copyright 2000 Ogasawara Hiroyuki

#include	<zheaders.h>
#include	"zau_fs.h"

#ifndef	TRUE
# define	TRUE	1
# define	FALSE	0
#endif


static char *
_StrCpyHi( char *str, char *ptr )
{
	for(; *ptr ;){
		char	ch= *ptr++;
		if( ch >= 'a' && ch <= 'z' ){
			ch^= 0x20;
		}
		*str++= ch;
	}
	*str= '\0';
	return	str;
}

static void
_StrCpyLow( char *str, char *ptr )
{
	for(; *ptr ;){
		char	ch= *ptr++;
		if( ch >= 'A' && ch <= 'Z' ){
			ch^= 0x20;
		}
		*str++= ch;
	}
	*str= '\0';
}


static int
_StrLen( char *ptr )
{
	int	i;
	for( i= 0 ; *ptr++ ; i++ );
	return	i;
}

static void
_ZauName( char *str, char *drv, char *name )
{
	str= _StrCpyHi( str, drv );
	_StrCpyHi( str, name );
}


int
_ZauReadDir( char *drv, int *no, char *name )
{
	char	pathbuf[24];
	char	namebuf[24];
	FSEARCHMODE	pmode;
	pmode.DirNo= *no;
	pmode.Behav= FC_SEARCH_INV_NEXT;
	_ZauName( pathbuf, drv, "*.*" );
	if( FileSearch( pathbuf, &pmode, namebuf ) < 0 ){
		return	FALSE;
	}
	_StrCpyLow( name, namebuf+3 );
	*no= pmode.DirNo;
	return	TRUE;
}

int
_ZauLookup( char *drv, char *name, struct __zaurus_dir_ent *ent )
{
	FDIRBUF	dirp;
	char	namebuf[24];
	_ZauName( namebuf, drv, name );
	_StrCpyHi( ent->fname, name );
	ent->namelen= _StrLen( ent->fname );
	if( FileDirAcc( namebuf, FC_ACC_READ, &dirp ) < 0 ){
		return	FALSE;
	}
	ent->file_length= dirp.Size;
	ent->mode= 0777;
	if( dirp.Atr & (PROTECT_FILE|SYSTEM_FILE) ){
		ent->mode= 0555;
	}
	ent->mode|= S_IFREG;
	{
		unsigned short	d= dirp.Date;
		// 0000000 0000 00000
		int	day=	d & 31;
		int	month=	(d>>5) & 15;
		int	year=	(d>>9) + 1980 - 1970;

		unsigned short	t= dirp.Time;
		// 00000 000000 00000
		int	sec=	(t<<1) & 62;
		int	minute=	(t>>5) & 63;
		int	hour=	(t>>11);

		static unsigned short	_mdays[]= {
		0,
		0,
		31,		// 1
		31+28,		// 2
		31+28+31,		// 3
		31+28+31+30,		// 4
		31+28+31+30+31,		// 5
		31+28+31+30+31+30,		// 6
		31+28+31+30+31+30+31,		// 7
		31+28+31+30+31+30+31+31,		// 8
		31+28+31+30+31+30+31+31+30,		// 9
		31+28+31+30+31+30+31+31+30+31,		// 10
		31+28+31+30+31+30+31+31+30+31+30,		// 11
		//31+28+31+30+31+30+31+31+30+31+30+31,		// 12
		};

		day+= year*365	+ ((year+2)/4)
				- ((year+70)/100)
				+ ((year+370)/400)
				+ _mdays[month] -1;
		if( (!((year+2)&3)) && month < 2 )
			day--;
		ent->time= day * 24*60*60 +
			hour * 60*60 + minute * 60 + sec;
	}

//	ent->time= 0;
	return	TRUE;
}


int
_ZauOpen( char *drv, char *name, int mode )
{
	int	fh;
	char	namebuf[24];
	static FOPENMODE	_Mode[]= { R_OPEN, W_OPEN, RW_OPEN };
	_ZauName( namebuf, drv, name );

	if( (fh= (int)FileOpen( namebuf, _Mode[mode] )) < 0 ){
		return	-1;
	}
	return	fh;

}

int
_ZauCreate( char *drv, char *name )
{
	int	fh;
	char	namebuf[24];
	_ZauName( namebuf, drv, name );

	if( (fh= (int)FileCreate( namebuf, USUAL_FILE )) < 0 ){
		return	-1;
	}
	return	fh;

}


int
_ZauClose( int fh )
{
	return	FileClose( (FILEHD)fh );
}

int
_ZauRead( int fh, char *buf, int len )
{
	unsigned int	ilen= len;
	if( FileBlockRead( (FILEHD)fh, FC_FILE_END_PH,
				(unsigned char*)buf, &ilen ) < 0 ){
		return	-1;
	}
	return	ilen;
}

int
_ZauSeek( int fh, int pos )
{
	return	FileMovePtr( (FILEHD)fh, FC_FROM_FILETOP, pos );
}

int
_ZauWrite( int fh, char *buf, int len )
{
	unsigned int	ilen= len;
	if( FileBlockWrite( (FILEHD)fh, (unsigned char*)buf, &ilen ) < 0 ){
		return	-1;
	}
	return	ilen;
}


int
_ZauRename( char *drv, char *oldname, char *newname )
{
	char	namebuf[24];
	char	newbuf[24];
	_ZauName( namebuf, drv, oldname );
	_StrCpyHi( newbuf, newname );

	return	FileRename( namebuf, newbuf );
}



int
_ZauDelete( char *drv, char *name )
{
	char	namebuf[24];
	_ZauName( namebuf, drv, name );

	return	FileDelete( namebuf );
}


