//	Copyright 2000 Ogasawara Hiroyuki

//	ZaurusFileSystem


#include	<linux/sched.h>
#include	<linux/kernel.h>
#include	<linux/errno.h>
#include	<linux/stat.h>
#include	<linux/init.h>
#include	<linux/file.h>
#include	<linux/locks.h>
#include	<asm/uaccess.h>
#include	<linux/zau_fs.h>

//#include	"../../_debugcode.h"
#define		_DebugOutput	//
//----------------------------------------------------------------------------
extern int	_ZauReadDir( char *drv, int *no, char *name );
extern int	_ZauLookup( char *drv, char *name,struct __zaurus_dir_ent*ent);
extern int	_ZauOpen( char *drv, char *name, int mode );
extern int	_ZauCreate( char *drv, char *name );
extern int	_ZauClose( int fh );
extern int	_ZauRead( int fh, char *buf, int len );
extern int	_ZauWrite( int fh, char *buf, int len );
extern int	_ZauSeek( int fh, int pos );
extern int	_ZauRename( char *drv, char *oldname, char *newname );
extern int	_ZauDelete( char *drv, char *name );
//----------------------------------------------------------------------------

#define		_ZauDirPath(sb)		((char*)(&(sb)->u.ext2_sb))
#define		_ZauDirFile(inode)	((char*)(&(inode)->u.ext2_i))
#define		_ZauDirFH(file)		((int)((file)->private_data))
#define		_ZauDirSetFH(file, fh)	(((file)->private_data)=(void*)(fh))


struct inode *
zau_get_inode( struct super_block *sb, struct __zaurus_dir_ent *ent )
{
	struct inode	*inode;
	inode= get_empty_inode();
	if( !inode ){
		return	NULL;
	}
	inode->i_sb= sb;
	inode->i_dev= sb->s_dev;
//	inode->i_ino= ent->inode;
	inode->i_mode= ent->mode;
	inode->i_size= ent->file_length;
	inode->i_uid= 0;
	inode->i_gid= 0;
	inode->i_op= ent->i_op;
	inode->i_nlink= 1;
	inode->i_atime= ent->time;
	inode->i_mtime= ent->time;
	inode->i_ctime= ent->time;
	strcpy( _ZauDirFile(inode), ent->fname );
	return	inode;
}

//---------------------------------------------------------------------------

static const int	_loadbufsize= __ZAUFS_IO_PAGESIZE;
static char		_loadbuf[__ZAUFS_IO_PAGESIZE];

#if 0
static int
zau_file_lseek( struct file *file, loff_t offset, int mode )
{
_DebugOutput( "zau_file_lseek %d", _ZauDirFH(file) );
	return	0;
}
#endif

static ssize_t
zau_file_read( struct file *file, char *buf, size_t count, loff_t *ppos )
{
	int	totallen;
	int	fh= _ZauDirFH( file );
//_DebugOutput( "zau_file_read %d %d", fh, count );
	if( !count ){
		return	0;
	}
	if( !buf || count < 0 ){
		return	-EINVAL;
	}
	_ZauSeek( fh, *ppos );
	totallen= 0;
	do{
		int	readlen= count > _loadbufsize ? _loadbufsize : count;
		int	copylen= _ZauRead( fh, _loadbuf, readlen );
		if( copylen <= 0 )
			break;
		copy_to_user( buf, _loadbuf, copylen );
		buf+= copylen;
		count-= copylen;
		totallen+= copylen;
	}while( count > 0 );
	*ppos= *ppos + totallen;
//_DebugOutput( "zau_file_read totallen=%d", totallen );
	return	totallen;
}

static ssize_t
zau_file_write( struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	int	totallen;
	int	fh= _ZauDirFH( file );
	struct inode	*inode= file->f_dentry->d_inode;
	loff_t	pos;
//_DebugOutput( "zau_file_write %d %d", fh, count );
	if( !count ){
		return	0;
	}
	if( !buf || count < 0 ){
		return	-EINVAL;
	}
	if( file->f_flags & O_APPEND ){
		pos= inode->i_size;
	}else{
		pos= *ppos;
	}
	_ZauSeek( fh, pos );

	totallen= 0;
	do{
		int	writelen= count < _loadbufsize ? count : _loadbufsize;
		int	copylen;
		copy_from_user( _loadbuf, buf, writelen );
		copylen= _ZauWrite( fh, _loadbuf, writelen );
		if( copylen <= 0 )
			break;
		buf+= copylen;
		count-= copylen;
		totallen+= copylen;
	}while( count > 0 );
	*ppos= pos+= totallen;
	if( pos > inode->i_size ){
		inode->i_size= pos;
	}
	inode->i_mtime= inode->i_ctime= CURRENT_TIME;
	return	totallen;
}

static int
zau_file_open( struct inode *inode, struct file *file )
{
	int	fh;
	// 241
	// 0 8 64 512 = TRUNC
	// 1= WRITE
	// 64= CRE
	// 577 - 512 + 64 + 1
	//       200   40   1
//_DebugOutput( "zau-open mode(%x)", file->f_flags );
	if( file->f_flags & (O_CREAT|O_TRUNC) ){
		fh= _ZauCreate(
			_ZauDirPath( inode->i_sb ),
			_ZauDirFile( inode )
			);
	}else{
		fh= _ZauOpen(
			_ZauDirPath( inode->i_sb ),
			_ZauDirFile( inode ),
			(file->f_flags & O_ACCMODE)  );
	}
_DebugOutput( "zau_file_open [%s] %d mode=%x flag=%x",
			_ZauDirFile(inode), fh, file->f_mode, file->f_flags );
	if( fh < 0 ){
_DebugOutput( "zau_open err" );
	}
	_ZauDirSetFH( file, fh );
	return	0;
}

static int
zau_file_release( struct inode *inode, struct file *file )
{
//_DebugOutput( "zau_file_release %d", _ZauDirFH(file) );
	_ZauClose( _ZauDirFH(file) );
	return	0;
}


static int
zau_file_mmap( struct file *file, struct vm_area_struct *vma )
{
//_DebugOutput( "zau_file_mmap %d", _ZauDirFH(file) );
	return	0;
}

//----------------------------------------------------------------------------

#if 0
static int
zau_file_readpage( struct file *file, struct page *page )
{
//	char		*buffer= (char*)page_address( page );
//	unsigned long	offset= page->offset;

_DebugOutput( "zau_file_readpage" );

	return	0;
}

static int
zau_file_writepage( struct file *file, struct page *page )
{
_DebugOutput( "zau_file_writepage" );
	return	0;
}


static int
zau_file_get_block( struct inode *inode, long block,
		struct buffer_head *bh_result, int create )
{
_DebugOutput( "zau_file_get_block" );
	return	0;
}
#endif



#if 0
static int
zau_file_rename(
	struct inode *old_inode, struct dentry *old_dentry,
	struct inode *new_inode, struct dentry *new_dentry )
{
_DebugOutput( "zau_file_rename" );
	return	0;
}


static int
zau_file_create( struct inode *dir, struct dentry *old_dentry, int mode )
{
_DebugOutput( "zau_file_create" );
	return	0;
}

static int
zau_file_lookup()
{
_DebugOutput( "zau_file_lookup" );
	return	0;
}

static int
zau_file_revalidate()
{
_DebugOutput( "zau_file_revalidate" );
	return	0;
}

static int
zau_file_inode_revalidate()
{
_DebugOutput( "zau_file_inode_revalidate" );
	return	0;
}


#endif



//----------------------------------------------------------------------------

static struct file_operations zau_file_operations= {
	NULL,//zau_file_lseek,		/* lseek */
	zau_file_read,		/* read */
	zau_file_write,		/* write */
	NULL,			/* readdir */
	NULL,			/* poll */
	NULL,			/* ioctl */
	zau_file_mmap,		/* mmap */
	zau_file_open,		/* open */
	NULL,			/* flush */
	zau_file_release,	/* release */
	NULL,			/* fsync */
	NULL,			/* fasync */
	NULL,			/* check_media_change */
	NULL,//zau_file_revalidate,//NULL,		/* revalidate */
	NULL,			/* lock */
};

static struct inode_operations zau_file_inode_operations= {
	&zau_file_operations,	/* default file operations */
	NULL,//zau_file_create,//NULL			/* create */
	NULL,//zau_file_lookup,//	NULL,			/* lookup */
	NULL,			/* link */
	NULL,//zau_file_unlink,	/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,//zau_file_rename,//NULL	/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,//zau_file_get_block,	/* get_block */
	NULL,//zau_file_readpage,	/* readpage */
	NULL,//zau_file_writepage,	/* writepage */
	NULL,			/* flushpage */
	NULL,			/* truncate */
	NULL,			/* permission */
	NULL,			/* smap */
	NULL,//zau_file_inode_revalidate,//NULL,	/* revalidate */
};

//-------------------------------------------------------------------------
#if 0
static int
dop_zau_lookup_validate( struct dentry *dentry, int flags )
{
_DebugOutput( "dop_zau_lookup_validate" );
}

static int
dop_zau_hash_dentry( struct dentry *dentry, struct qstr *this )
{
_DebugOutput( "dop_zau_hash_dentry" );
}

static int
dop_zau_compare_dentry( struct dentry *dentry, struct qstr *a, struct qstr *b )
{
_DebugOutput( "dop_zau_compare_dentry" );
}

static void
dop_zau_delete_dentry( struct dentry *dentry )
{
_DebugOutput( "dop_zau_delete_dentry" );
}

static struct dentry_operations	zaufs_dentry_operations= {
	dop_zau_lookup_validate,	// d_revalidate(dentry*,int)
	dop_zau_hash_dentry,		// d_hash(dentry*,qstr*)
	dop_zau_compare_dentry,		// d_compare(dentry*,qstr*,qstr*)
	dop_zau_delete_dentry,		// d_delete(dentry*)
	dop_zau_release_dentry,		// d_release(dentry*)
	dop_zau_iput_dentry,		// d_iput(dentry*,inode*)
};
#endif

//-------------------------------------------------------------------------


static int
zau_root_readdir( struct file *filp, void *dirent, filldir_t filldir )
{
	unsigned int	ino;
	int		i;
	struct inode	*inode= filp->f_dentry->d_inode;
//_DebugOutput( "zau_root_readdir" );
	ino= inode->i_ino;
	i= filp->f_pos;
	switch( i ){
	case 0:
		if( filldir( dirent, ".", 1, i, ino ) < 0 ){
			return	0;
		}
		i++;
		filp->f_pos++;
	case 1:
		if( filldir( dirent, "..", 2, i, ino++ ) < 0 ){
			return	0;
		}
		i++;
		filp->f_pos++;
	default:
		{
			char	namebuf[24];
			int	no= filp->f_pos-2;
			for(; _ZauReadDir( _ZauDirPath(inode->i_sb), &no,
								namebuf ) ;){
				if( filldir( dirent, namebuf, strlen(namebuf),
						filp->f_pos, ino++ ) < 0 ){
					return	0;
				}
				filp->f_pos++;
				no++;
			}
		}
		break;
	}

	return	1;
}


static struct dentry *
zau_root_lookup( struct inode *inode, struct dentry *dentry )
{
	char	namebuf[80];
	struct __zaurus_dir_ent	ent;
	struct inode	*finode;

	memcpy( namebuf, dentry->d_name.name, dentry->d_name.len );
	namebuf[dentry->d_name.len]= '\0';
_DebugOutput( "zau_root_lookup %s", namebuf );

	if( _ZauLookup(
			_ZauDirPath( inode->i_sb ),
			namebuf, &ent ) ){
		ent.i_op= &zau_file_inode_operations;
		finode= zau_get_inode( inode->i_sb, &ent );
		if( finode ){
			d_add( dentry, finode );
			return	NULL;
		}
		return	ERR_PTR(-EINVAL);
	}
//	return	ERR_PTR(-ENOENT);
//	dentry->d_op= &zaufs_dentry_operations;
	d_add( dentry, NULL );
	return	NULL;
}

//-------------------------------------------------------------------------


static int
zau_root_unlink( struct inode *dir, struct dentry *dentry )
{
	char	namebuf[__ZAUFS_MAX_PATH_BUF];
	struct inode	*inode;
	memcpy( namebuf, dentry->d_name.name, dentry->d_name.len );
	namebuf[dentry->d_name.len]= '\0';
//_DebugOutput( "zau_root_unlink %s", namebuf );
	if( _ZauDelete( _ZauDirPath(dir->i_sb), namebuf ) < 0 ){
		return	-ENOENT;
	}
	inode= dentry->d_inode;
	inode->i_nlink--;
	d_delete( dentry );
	iput( inode );
	return	0;
}


static int
zau_root_rename(
	struct inode *old_inode_dir, struct dentry *old_dentry,
	struct inode *new_inode_dir, struct dentry *new_dentry )
{
	char	namebuf[__ZAUFS_MAX_PATH_BUF];
	char	namebuf2[__ZAUFS_MAX_PATH_BUF];
_DebugOutput( "zau_root_rename" );

	memcpy( namebuf, old_dentry->d_name.name, old_dentry->d_name.len );
	namebuf[old_dentry->d_name.len]= '\0';
	memcpy( namebuf2, new_dentry->d_name.name, new_dentry->d_name.len );
	namebuf2[new_dentry->d_name.len]= '\0';
	if( new_dentry->d_inode ){
		_ZauDelete( _ZauDirPath( old_inode_dir->i_sb ), namebuf2 );
	}
	if( _ZauRename(
			_ZauDirPath( old_inode_dir->i_sb ),
			namebuf, namebuf2 ) < 0 ){
		return	-ENOENT;
	}
//	d_instantiate( new_dentry, old_dentry->d_inode );
//	d_delete( old_dentry );

	return	0;
}


static int
zau_root_create( struct inode *dir, struct dentry *dentry, int mode )
{
	char		namebuf[__ZAUFS_MAX_PATH_BUF];
	struct inode	*finode;
	struct __zaurus_dir_ent	ent;
	int	fh;
	memcpy( namebuf, dentry->d_name.name, dentry->d_name.len );
	namebuf[dentry->d_name.len]= '\0';
_DebugOutput( "zau_root_create %s", namebuf );
	fh= _ZauCreate( _ZauDirPath(dir->i_sb), namebuf );
	_ZauClose( fh );
	_ZauLookup( _ZauDirPath(dir->i_sb), namebuf, &ent );
	ent.i_op= &zau_file_inode_operations;
	finode= zau_get_inode( dir->i_sb, &ent );
	d_instantiate( dentry, finode );
//	d_add( dentry, finode );
	return	0;
}

#if 0
static int
zau_revalidate_inode( struct dentry *dentry )
{
_DebugOutput( "zau_revalidate_inode" );
	return	0;
}
#endif


//-------------------------------------------------------------------------

static struct file_operations zau_root_operations= {
	NULL,			/* lseek */
	NULL,			/* read */
	NULL,			/* write */
	zau_root_readdir,	/* readdir */
	NULL,			/* poll */
	NULL,			/* ioctl */
	NULL,			/* mmap */
	NULL,			/* open */
	NULL,			/* flush */
	NULL,			/* release */
	NULL,			/* fsync */
	NULL,			/* fasync */
	NULL,			/* check_media_change */
	NULL,			/* revalidate */
	NULL,			/* lock */
};

static struct inode_operations zau_root_inode_operations= {
	&zau_root_operations,	/* default dir operations */
	zau_root_create,	/* create */
	zau_root_lookup,	/* lookup */
	NULL,			/* link */
	zau_root_unlink,	/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	zau_root_rename,	/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* get_block */
	NULL,			/* readpage */
	NULL,			/* writepage */
	NULL,			/* flushpage */
	NULL,			/* truncate */
	NULL,			/* permission */
	NULL,			/* smap */
NULL,//zau_revalidate_inode,//	NULL,			/* revalidate */
};

static void
zau_read_inode( struct inode *inode )
{
// from disk
//_DebugOutput( "zau_read_inode" );
}

static void
zau_write_inode( struct inode *inode )
{
// to disk
//_DebugOutput( "zau_write_inode" );
}


static void
zau_put_inode( struct inode *inode )
{
//_DebugOutput( "zau_put_inode" );
	if( inode->i_count == 1 ){
		inode->i_nlink= 0;
	}
}

static void
zau_delete_inode( struct inode *inode )
{
//_DebugOutput( "zau_delete_inode" );
	clear_inode( inode );
}

static int
zau_statfs( struct super_block *sb, struct statfs *buf, int bufsize )
{
_DebugOutput( "zau_statfs" );
#if 0
	struct statfs	attr;
	memset( &attr, 0, sizeof(struct statfs) );
	attr.f_type= ZAU_SUPER_MAGIC;
	attr.f_files= -1;
	attr.f_ffree= -1;
	attr.f_namelen= 12;
#endif
	return	0;
}

static void
zau_put_super( struct super_block *sb )
{
_DebugOutput( "zau_put_super" );
}

//---------
static int
zau_remount_filesystem( struct super_block *sb, int *ip, char *cp )
{
_DebugOutput( "zau_remount_filesystem" );
	return	0;
}

static void
zau_write_superblock( struct super_block *sb )
{
_DebugOutput( "zau_write_superblock" );
}

static struct super_operations	zau_sops= {
	zau_read_inode,		/* read inode */
	zau_write_inode,	/* write inode */
	zau_put_inode,		/* put inode */
	zau_delete_inode,	/* delete inode */
	NULL,//zau_notify_change,	/* notify change */
	zau_put_super,		/* put superblock */
zau_write_superblock,//	NULL,			/* write superblock */
	zau_statfs,		/* stat filesystem */
zau_remount_filesystem,//NULL			/* remount filesystem */
};

static struct __zaurus_dir_ent	_zaufs_root= {
	9, 5, "zaufs", 0,
	0777|S_IFDIR,
	0,	// len
	0,	// time
	&zau_root_inode_operations,
};


struct super_block *
zau_read_super( struct super_block *sb, void *raw_data, int silent )
{
	struct inode	*root_inode;
	char		*ZauDir= "F1:";
	char		DirBuf[80];

	if( raw_data ){
		char	*p= (char*)raw_data;
		char	*s= DirBuf;
		for(; *p ; p++ ){
			if( *p == '=' ){
				for( p++ ; *p && *p != ',';){
					char	ch= *p++;
					if( ch == '/' )
						ch= '\\';
					*s++= ch;
				}
				break;
			}
		}
		*s= '\0';
		ZauDir= DirBuf;
	}

	lock_super( sb );
	sb->s_blocksize= 1024;
	sb->s_blocksize_bits= 10;
	sb->s_magic= ZAU_SUPER_MAGIC;
	sb->s_flags= 0;
	sb->s_op= &zau_sops;
	strcpy( _ZauDirPath(sb), ZauDir );
	if( !(root_inode= zau_get_inode( sb, &_zaufs_root )) ){
		goto	_out_no_root;
	}
	if( !(sb->s_root= d_alloc_root( root_inode )) ){
		goto	_out_no_root;
	}
	unlock_super( sb );
	return	sb;

_out_no_root:
	iput( root_inode );
	sb->s_dev= 0;
	unlock_super( sb );
	return	NULL;

}


static struct file_system_type zau_fs_type= {
	"zaufs",
	0,
	zau_read_super,
	NULL
};


int __init init_zau_fs()
{
//_DebugOutput( "register_filesys" );
	return	register_filesystem( &zau_fs_type );
}

