// 2000/10/23 Ogasawara Hiroyuki

#include	<linux/config.h>
#include	<linux/errno.h>
#include	<linux/types.h>
#include	<linux/socket.h>
#include	<linux/in.h>
#include	<linux/kernel.h>
#include	<linux/major.h>
#include	<linux/sched.h>
#include	<linux/timer.h>
#include	<linux/string.h>
#include	<linux/sockios.h>
#include	<linux/net.h>
#include	<linux/fcntl.h>
#include	<linux/mm.h>
#include	<linux/interrupt.h>
#include	<linux/proc_fs.h>
#include	<linux/stat.h>
#include	<linux/init.h>
#include	<linux/poll.h>

#include	<asm/uaccess.h>
#include	<asm/system.h>

#include	<linux/smp_lock.h>
#include	<linux/inet.h>
#include	<linux/netdevice.h>
#include	<net/ip.h>
#include	<net/protocol.h>
#include	<net/arp.h>
#include	<net/route.h>
#include	<net/tcp.h>
#include	<net/udp.h>
#include	<net/sock.h>
#include	<net/inet_common.h>

#define		ZAPI	/*ZaurusAPI*/
#define		ErrorZaurusAPItoLinux(n)	(n)

//----------------------------------------------------------------------------
// Zaurus API Call
typedef		void	*SADR;
extern int	socket( int af, int type, int protocol, int *err );
extern int	accept( int fd, SADR*, int *len, int *err );
extern int	listen( int fd, int backlog );
extern int	nselect( void *selp, int cnt, int *waitp, int mode,
				unsigned long arg, int *err );
extern int	bind( int fd, SADR*, int len );
extern int	getpeername( int fd, SADR*, int *len );
extern int	getsockname( int fd, SADR*, int *len );
extern int	connect( int fd, SADR*, int len );
extern int	sendto( int fd, char *buf, int len, int flags, SADR *hd,
			int hdlen, int *err );
extern int	recvfrom( int fd, char *buf, int len, int flags, SADR *hd,
			int *hdlen, int *err );
extern int	close( int fd );
extern int	shutdown( int fd, int direction );
//----------------------------------------------------------------------------

static int
_ConvertMsgFlags( int flags )
{
	int	zflags= flags & (MSG_OOB|MSG_PEEK|MSG_DONTROUTE);
	if( flags & MSG_DONTWAIT ){
		zflags|= 0x4000;	// MSG_NONBLOCKING
	}
	if( flags & MSG_WAITALL ){
		zflags|= 0x8000;	// MSG_BLOCKING
	}
	return	zflags;
}

int
inet_release( struct socket *sock )
{
	int	err;
	err= ZAPI close( (int)sock->sk );
	err= ErrorZaurusAPItoLinux( err );
	return	err;
}


static int
inet_bind( struct socket *sock, struct sockaddr *uaddr, int addr_len )
{
	int	err;
	err= ZAPI bind( (int)sock->sk, (SADR*)uaddr, addr_len );
	err= ErrorZaurusAPItoLinux( err );
	return	err;
}

static int
inet_connect( struct socket *sock, struct sockaddr *uaddr,
					int addr_len, int flags )
{
	int	err;
	err= ZAPI connect( (int)sock->sk, (SADR*)uaddr, addr_len );
	err= ErrorZaurusAPItoLinux( err );
	return	err;
}

int
inet_accept( struct socket *sock, struct socket *newsock, int flags )
{
	int	err;
	int	zerr;
	struct sockaddr	newaddr;
	int	addr_len= sizeof(struct sockaddr);
	err= ZAPI accept( (int)sock->sk, (SADR*)&newaddr, &addr_len, &zerr );
	if( err < 0 ){
		err= ErrorZaurusAPItoLinux( err );
		return	err;
	}else{
		newsock->sk= (struct sock*)err;
		newsock->state= SS_CONNECTED;
	}
	return	0;
}

static int
inet_getname( struct socket *sock, struct sockaddr *uaddr, int *uaddr_len,
						int peer )
{
	int	err;
	if( peer ){
		err= ZAPI getpeername( (int)sock->sk,
						(SADR*)uaddr, uaddr_len );
	}else{
		err= ZAPI getsockname( (int)sock->sk,
						(SADR*)uaddr, uaddr_len );
	}
	err= ErrorZaurusAPItoLinux( err );
	return	err;
}

#define	__MAX_IOBUFFER		(1024*2)

int
inet_recvmsg( struct socket *sock, struct msghdr *msg, int size,
				int flags, struct scm_cookie *scm )
{
	int	err;
	int	zerr;
	int	rlen;
	unsigned char	__NetworkIO_Buffer[__MAX_IOBUFFER];
	if( msg->msg_iovlen <= 0 )
		return	0;
	if( size > __MAX_IOBUFFER )
		size= __MAX_IOBUFFER;
	if( sock->type == SOCK_STREAM ){
		err= ZAPI recv( (int)sock->sk,
			(char*)__NetworkIO_Buffer,
			//(char*)msg->msg_iov->iov_base,
			size,//msg->msg_iov->iov_len,
			_ConvertMsgFlags( flags ),
			&zerr );
	}else{
		err= ZAPI recvfrom( (int)sock->sk,
			(char*)__NetworkIO_Buffer,
			//(char*)msg->msg_iov->iov_base,
			size,//msg->msg_iov->iov_len,
			_ConvertMsgFlags( flags ),
			(SADR*)msg->msg_name,
			&msg->msg_namelen, &zerr );
	}
	if( err < 0 ){
		err= ErrorZaurusAPItoLinux( err );
	}else{
		copy_to_user(
			(void*)msg->msg_iov->iov_base,
			(void*)__NetworkIO_Buffer,
				err );
	}
	return	err;
}

int
inet_sendmsg( struct socket *sock, struct msghdr *msg, int size,
				struct scm_cookie *scm )
{
	int	err;
	int	zerr;
	unsigned char	__NetworkIO_Buffer[__MAX_IOBUFFER];
	if( msg->msg_iovlen <= 0 )
		return	0;
	if( size > __MAX_IOBUFFER )
		size= __MAX_IOBUFFER;
	copy_from_user(
			(void*)__NetworkIO_Buffer,
			(void*)msg->msg_iov->iov_base,
				size );
	if( sock->type == SOCK_STREAM ){
		err= ZAPI send( (int)sock->sk,
			__NetworkIO_Buffer,
			//(char*)msg->msg_iov->iov_base,
			size,//msg->msg_iov->iov_len,
			_ConvertMsgFlags( msg->msg_flags ),
			&zerr );
	}else{
		err= ZAPI sendto( (int)sock->sk,
			__NetworkIO_Buffer,
			//(char*)msg->msg_iov->iov_base,
			size,//msg->msg_iov->iov_len,
			_ConvertMsgFlags( msg->msg_flags ),
			(SADR*)msg->msg_name,
			msg->msg_namelen, &zerr );
	}
	err= ErrorZaurusAPItoLinux( err );
	return	err;
}

int
inet_shutdown( struct socket *sock, int how )
{
	int	err;
	err= ZAPI shutdown( (int)sock->sk, how );
	err= ErrorZaurusAPItoLinux( err );
	return	err;
}

unsigned int
inet_poll( struct file *file, struct socket *sock, poll_table *wait )
{
#if 0
	struct _sel {
		unsigned short	se_inflags;
		unsigned short	se_outflags;
		short		se_fd;
		short		se_lreserved;
		unsigned int	se_user;
		unsigned int	se_2reserved;
	} sel;
	int	zwait;
	int	zerr;
	sel.se_outflags= 0;
	sel.se_inflags= 0
		|1		// READ_NOTIFY
		//|2		// WRITE_NOTIFY
		//|4		// ACCEPT
		//|8		// CLOSE
		//|16		// CONNECT
		//|32		// EXCEPT
		//|64		// RSHUTDOWN
		//|128		// TIMEOUT
		//|256		// WSHUTDOWN
		//|512		// URGENT
		//|1024		// OOB
		;
	sel.se_fd= (short)sock->sk;
	ZAPI nselect( (void*)&sel, 1, &zwait, 0, 0, &zerr );
#endif
	return	0;
}

int
inet_listen( struct socket *sock, int backlog )
{
	int	err;
	err= ZAPI listen( (int)sock->sk, backlog );
	err= ErrorZaurusAPItoLinux( err );
	return	err;
}



static struct proto_ops	inet_stream_ops= {
	PF_INET,
	inet_release,
	inet_bind,
	inet_connect,
	sock_no_socketpair,
	inet_accept,
	inet_getname, 
	inet_poll,
	sock_no_ioctl,
	inet_listen,
	inet_shutdown,
	sock_no_setsockopt,
	sock_no_getsockopt,
	sock_no_fcntl,
	inet_sendmsg,
	inet_recvmsg,
	sock_no_mmap
};

static struct proto_ops	inet_dgram_ops= {
	PF_INET,
	inet_release,
	inet_bind,
	inet_connect,
	sock_no_socketpair,
	sock_no_accept,
	inet_getname, 
	inet_poll,
	sock_no_ioctl,
	sock_no_listen,
	inet_shutdown,
	sock_no_setsockopt,
	sock_no_getsockopt,
	sock_no_fcntl,
	inet_sendmsg,
	inet_recvmsg,
	sock_no_mmap
};

/*
		ZaurusAPI	Linux
------------------------------------------------------
SOCK_STREAM	1		1
SOCK_DGRAM	2		2
SOCK_RAW	3		3
SOCK_RDM	4		4
SOCK_SEQPACKET	5		5
SOCK_PACK_EX	6		-
SOCK_PACKET	-		10

		ZaurusAPI	Linux
------------------------------------------------------
AF_UNSPEC	0
AF_INTRA	1
AF_INET		2
AF_NS		3
AF_NBS		4
AF_CCITT	5
AF_SNA		6
AF_ARP		7
AF_KDB		8
AF_RARP		9

AF_UNSPEC			0
AF_UNIX				1
AF_LOCAL			1
AF_INET				2
AF_AX25				3
AF_INET6			10

MSG_OOB		1		1
MSG_PEEK	2		2
MSG_DONTROUTE	4		4
MSG_NONBLOCKING	0x4000
MSG_BLOCKING	0x8000

MSG_DONTWAIT			0x040
MSG_WAITALL			0x100

*/


static int
inet_create( struct socket *sock, int protocol )
{
	int	err;
	int	zerr;

	err= ZAPI socket( /*family*/ AF_INET,
			sock->type, protocol, &zerr );
	if( err >= 0 ){
		sock->sk= (struct sock*)err;
		err= 0;
	}else{
		err= ErrorZaurusAPItoLinux(err);
	}

	switch( sock->type ){
	case SOCK_SEQPACKET:
	case SOCK_RAW:
	case SOCK_STREAM:
		sock->ops= &inet_stream_ops;
		break;
	case SOCK_DGRAM:
		sock->ops= &inet_dgram_ops;
		break;
	default:
		return	-EINVAL;
	}

	return	err;
}



static struct net_proto_family	zaunet_family_ops= {
	PF_INET,
	inet_create
};




void __init
zaunet_proto_init( struct net_proto *pro )
{
	printk( KERN_INFO "ZAUNET: Zaurus TCP/IP\n");

  	(void)sock_register( &zaunet_family_ops );
}

